#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## minizinc_text.py
##
"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        MiniZincText

    ==================
    Module description
    ==================

    ==============
    Module details
    ==============
"""

from ..transformations.get_variables import get_variables
from .solver_interface import SolverInterface
from ..expressions import *
from ..expressions.variables import _NumVarImpl, _IntVarImpl, _BoolVarImpl, NegBoolView
import numpy as np

# translate expression tree to MiniZinc textual model
class MiniZincText(SolverInterface):
    # does not do solving itself (you can subclass it)
    # does provide conversion to text model
    def __init__(self):
        self.name = "minizinc_text"

    def supported(self):
        return True # always possible

    def solve(self, model):
        print(self.convert(model))

    def convert(self, model):
        modelvars = get_variables(model)
        txt_vars = "% Generated by CPMpy\ninclude \"globals.mzn\";\n\n"
        for var in modelvars:
            if isinstance(var, _BoolVarImpl):
                txt_vars += "var bool: BV{};\n".format(var.name)
            elif isinstance(var, _IntVarImpl):
                txt_vars += "var {}..{}: IV{};\n".format(var.lb, var.ub, var.name)

        txt_cons = ""
        for con in model.constraints:
            txt_cons += "constraint {};\n".format(self.convert_expression(con))

        txt_obj = "solve "
        if model.objective is None:
            txt_obj += "satisfy;"
        else:
            if model.objective_max:
                txt_obj += "maximize "
            else:
                txt_obj += "minimize "
            txt_obj += "{};\n".format(self.convert_expression(model.objective))
                
        return txt_vars+"\n"+txt_cons+txt_obj

    # expected to return a string
    def convert_expression(self, expr):
        if is_any_list(expr):
            if isinstance(expr, np.ndarray):
                # must flatten
                expr_str = [self.convert_expression(e) for e in expr.flat]
            else:
                expr_str = [self.convert_expression(e) for e in expr]
            return "[{}]".format(",".join(expr_str))

        if not isinstance(expr, Expression) or \
           isinstance(expr, _NumVarImpl):
            if expr is True:
                return "true"
            if expr is False:
                return "false"
            # default
            return str(expr)
        
        args_str = [self.convert_expression(e) for e in expr.args]

        # standard expressions: comparison, operator, element
        if isinstance(expr, Comparison):
            # pretty printing: add () if nested comp/op
            for e in expr.args:
                if isinstance(e, (Comparison,Operator)):
                    for i in [0,1]:
                        args_str[i] = "({})".format(args_str[i])
            # infix notation
            return "{} {} {}".format(args_str[0], expr.name, args_str[1])

        if isinstance(expr, Operator):
            # some names differently (the infix names!)
            printmap = {'and': '/\\', 'or': '\\/',
                        'sum': '+', 'sub': '-',
                        'mul': '*', 'div': '/', 'pow': '^'}
            op_str = expr.name
            if op_str in printmap:
                op_str = printmap[op_str]

            # TODO: pretty printing of () as in Operator?

            # special case: unary -
            if self.name == '-':
                return "-{}".format(args_str[0])

            # special case, infix: two args
            if len(args_str) == 2:
                for i,arg_str in enumerate(args_str):
                    if isinstance(expr.args[i], Expression):
                        args_str[i] = "("+args_str[i]+")"
                return "{} {} {}".format(args_str[0], op_str, args_str[1])

            # special case: n-ary (non-binary): rename operator
            printnary = {'and': 'forall', 'or': 'exists', 'xor': 'xorall', 'sum': 'sum'}
            if expr.name in printnary:
                op_str = printnary[expr.name]
                return "{}([{}])".format(op_str, ",".join(args_str))

            # default: prefix printing
            return "{}({})".format(op_str, ",".join(args_str))

        if isinstance(expr, Element):
            subtype = "int"
            # TODO: need better bool check... is_bool() or type()?
            if all((v == 1) is v for v in iter(expr.args[0])):
                subtype = "bool"
            # minizinc is offset 1, which can be problematic for element...
            idx = args_str[1]
            if isinstance(expr.args[1], _IntVarImpl) and expr.args[1].lb == 0:
                idx = "{}+1".format(idx)
            # almost there
            txt  = "\n    let {{ array[int] of var {}: arr={} }} in\n".format(subtype, args_str[0])
            txt += f"      arr[{idx}]"
            return txt
        
        # rest: global constraints
        if expr.name.endswith('circuit'): # circuit, subcircuit
            # minizinc is offset 1, which can be problematic here...
            if any(isinstance(e, _IntVarImpl) and e.lb == 0 for e in expr.args):
                # redo args_str[0]
                args_str = ["{}+1".format(self.convert_expression(e)) for e in expr.args]
        

        # default
        return "{}([{}])".format(expr.name, ",".join(args_str))



"""
    CPMpy interfaces to (the Python API interface of) solvers

    Solvers typically use some of the generic transformations in
    `transformations` as well as specific reformulations to map the
    CPMpy expression to the solver's Python API

    ==================
    List of submodules
    ==================
    .. autosummary::
        :nosignatures:

        ortools
        pysat
        utils

    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        CPM_ortools
        CPM_pysat

    =================
    List of functions
    =================
    .. autosummary::
        :nosignatures:

        param_combinations
"""

from .utils import builtin_solvers, get_supported_solvers, param_combinations
from .ortools import CPM_ortools
from .pysat import CPM_pysat
# from minizinc import CPMpyMiniZinc # closed for maintenance



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## globalconstraints.py
##
"""
    Global constraints conveniently express non-primitive constraints.

    Using global constraints
    ------------------------

    Solvers can have specialised implementations for global constraints. CPMpy has GlobalConstraint
    expressions so that they can be passed to the solver as is when supported.

    If a solver does not support a global constraint (see solvers/) then it will be automatically
    decomposed by calling its `.decompose()` function.

    As a user you **should almost never subclass GlobalConstraint()** unless you know of a solver that
    supports that specific global constraint, and that you will update its solver interface to support it.

    For all other use cases, it sufficies to write your own helper function that immediately returns the
    decomposition, e.g.:

    .. code-block:: python

        def alldifferent_except0(args):
            return [ ((var1!= 0) & (var2 != 0)).implies(var1 != var2) for var1, var2 in all_pairs(args)]


    Numeric global constraints
    --------------------------

    CPMpy also implements __Numeric Global Constraints__. For these, the CPMpy GlobalConstraint does not
    exactly match what is implemented in the solver, but for good reason!!

    For example solvers may implement the global constraint `Minimum(iv1, iv2, iv3) == iv4` through an API
    call `addMinimumEquals([iv1,iv2,iv3], iv4)`.

    However, CPMpy also wishes to support the expressions `Minimum(iv1, iv2, iv3) > iv4` as well as
    `iv4 + Minimum(iv1, iv2, iv3)`. 

    Hence, the CPMpy global constraint only captures the `Minimum(iv1, iv2, iv3)` part, whose return type
    is numeric and can be used in any other CPMpy expression. Only at the time of transforming the CPMpy
    model to the solver API, will the expressions be decomposed and auxiliary variables introduced as needed
    such that the solver only receives `Minimum(iv1, iv2, iv3) == ivX` expressions.
    This is the burden of the CPMpy framework, not of the user who wants to express a problem formulation.


    Subclassing GlobalConstraint
    ----------------------------
    
    If you do wish to add a GlobalConstraint, because it is supported by solvers or because you will do
    advanced analysis and rewriting on it, then preferably define it with a standard decomposition, e.g.:

    .. code-block:: python

        class my_global(GlobalConstraint):
            def __init__(self, args):
                super().__init__("my_global", args)

            def decompose(self):
                return [self.args[0] != self.args[1]] # your decomposition

    If it is a __numeric global constraint__ meaning that its return type is numeric (see `Minimum` and `Element`)
    then set `is_bool=False` in the super() constructor and preferably implement `.value()` accordingly.


    Alternative decompositions
    --------------------------
    
    For advanced use cases where you want to use another decomposition than the standard decomposition
    of a GlobalConstraint expression, you can overwrite the 'decompose' function of the class, e.g.:

    .. code-block:: python

        def my_circuit_decomp(self):
            return [self.args[0] == 1] # does not actually enforce circuit
        circuit.decompose = my_circuit_decomp # attach it, no brackets!

        vars = intvar(1,9, shape=10)
        constr = circuit(vars)

        Model(constr).solve()

    The above will use 'my_circuit_decomp', if the solver does not
    natively support 'circuit'.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        AllDifferent
        AllEqual
        Circuit
        Table
        Minimum
        Maximum
        Element

"""
import warnings # for deprecation warning
from .core import Expression
from .variables import boolvar, intvar, cpm_array
from .utils import flatlist, all_pairs, argval


# Base class GlobalConstraint

class GlobalConstraint(Expression):
    """
        Abstract superclass of GlobalConstraints

        Like all expressions it has a `.name` and `.args` property.
        Overwrites the `.is_bool()` method. You can indicate
        in the constructer whether it has Boolean return type or not.
    """
    # is_bool: whether this is normal constraint (True or False)
    #   not is_bool: it computes a numeric value (ex: Minimum, Element)
    def __init__(self, name, arg_list, is_bool=True):
        super().__init__(name, arg_list)
        self._is_bool = is_bool

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return self._is_bool

    def decompose(self):
        """
            Returns a decomposition into smaller constraints.

            The decomposition might create auxiliary variables
            and use other other global constraints as long as
            it does not create a circular dependency.
        """
        return None


# Global Constraints (with Boolean return type)


def alldifferent(args):
    warnings.warn("Deprecated, use AllDifferent(v1,v2,...,vn) instead, will be removed in stable version", DeprecationWarning)
    return AllDifferent(*args) # unfold list as individual arguments
class AllDifferent(GlobalConstraint):
    """All arguments have a different (distinct) value
    """
    def __init__(self, *args):
        super().__init__("alldifferent", flatlist(args))

    def decompose(self):
        """Returns the decomposition
        """
        return [var1 != var2 for var1, var2 in all_pairs(self.args)]


def allequal(args):
    warnings.warn("Deprecated, use AllEqual(v1,v2,...,vn) instead, will be removed in stable version", DeprecationWarning)
    return AllEqual(*args) # unfold list as individual arguments
class AllEqual(GlobalConstraint):
    """All arguments have the same value
    """
    def __init__(self, *args):
        super().__init__("allequal", flatlist(args))

    def decompose(self):
        """Returns the decomposition
        """
        return [var1 == var2 for var1, var2 in all_pairs(self.args)]

def circuit(args):
    warnings.warn("Deprecated, use Circuit(v1,v2,...,vn) instead, will be removed in stable version", DeprecationWarning)
    return Circuit(*args) # unfold list as individual arguments
class Circuit(GlobalConstraint):
    """The sequence of variables form a circuit, where x[i] = j means that j is the successor of i.
    """
    def __init__(self, *args):
        super().__init__("circuit", flatlist(args))

    def decompose(self):
        """
            Decomposition for Circuit

            Not sure where we got it from,
            MiniZinc has slightly different one:
            https://github.com/MiniZinc/libminizinc/blob/master/share/minizinc/std/fzn_circuit.mzn
        """
        succ = cpm_array(self.args)
        n = len(succ)
        order = intvar(0,n-1, shape=n)
        return [
            # different successors
            AllDifferent(succ),
            # different orders
            AllDifferent(order),
            # last one is '0'
            order[n-1] == 0,
            # loop: first one is successor of '0'
            order[0] == succ[0],
            # others: ith one is successor of i-1
        ] + [order[i] == succ[order[i-1]] for i in range(1,n)]

class Table(GlobalConstraint):
    """The values of the variables in 'array' correspond to a row in 'table'
    """
    def __init__(self, array, table):
        super().__init__("table", [array, table])

    def decompose(self):
        raise NotImplementedError("TODO: table decomposition")


# Numeric Global Constraints (with integer-valued return type)


class Minimum(GlobalConstraint):
    """
        Computes the minimum value of the arguments

        It is a 'functional' global constraint which implicitly returns a numeric variable
    """
    def __init__(self, arg_list):
        super().__init__("min", flatlist(arg_list), is_bool=False)

    def value(self):
        return min([argval(a) for a in self.args])

class Maximum(GlobalConstraint):
    """
        Computes the maximum value of the arguments

        It is a 'functional' global constraint which implicitly returns a numeric variable
    """
    def __init__(self, arg_list):
        super().__init__("max", flatlist(arg_list), is_bool=False)

    def value(self):
        return max([argval(a) for a in self.args])

def element(arg_list):
    warnings.warn("Deprecated, use Circuit(v1,v2,...,vn) instead, will be removed in stable version", DeprecationWarning)
    assert (len(arg_list) == 2), "Element expression takes 2 arguments: Arr, Idx"
    return Element(arg_list[0], arg_list[1])
class Element(GlobalConstraint):
    """
        The 'Element' global constraint enforces that the result equals Arr[Idx]
        with 'Arr' an array of constants of variables (the first argument)
        and 'Idx' an integer decision variable, representing the index into the array.

        Solvers implement it as Arr[Idx] == Y, but CPMpy will automatically derive or create
        an appropriate Y. Hence, you can write expressions like Arr[Idx] + 3 <= Y

        Element is a CPMpy built-in global constraint, so the class implements a few more
        extra things for convenience (.value() and .__repr__()). It is also an example of
        a 'numeric' global constraint.
    """

    def __init__(self, arr, idx):
        super().__init__("element", [arr, idx], is_bool=False)

    def value(self):
        idxval = argval(self.args[1])
        if not idxval is None:
            return argval(self.args[0][idxval])
        return None # default

    def __repr__(self):
        return "{}[{}]".format(self.args[0], self.args[1])




"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        SolverInterface
        SolverStatus
        ExitStatus

    ==================
    Module description
    ==================
    Contains the abstract class `SolverInterface` for defining solver interfaces,
    as well as a class `SolverStatus` that collects solver statistics,
    and the `ExitStatus` class that represents possible exist statuses.

    Each solver has its own class that inherits from `SolverInterface`.

"""
#
#==============================================================================
from enum import Enum
import time

#==============================================================================
class SolverInterface(object):
    """
        Abstract class for defining solver interfaces. All classes implementing
        the ``SolverInterface``
    """

    # REQUIRED functions:

    @staticmethod
    def supported():
        """
            Check for support in current system setup. Return True if the system
            has package installed or supports solver, else returns False.

        Returns:
            [bool]: Solver support by current system setup.
        """
        return False

    # REQUIRED functions to mimic `Model` interface:

    def __init__(self):
        self.cpm_status = SolverStatus("dummy") # status of solving this model

    def __add__(self):
        """
            Adds a constraint to the solver, eagerly (e.g. instantly passed to API)
        """
        raise NotImplementedError("Solver does not support eagerly adding constraints")

    def minimize(self, expr):
        """
            Minimize the given objective function

            `minimize()` can be called multiple times, only the last one is stored
        """
        raise NotImplementedError("Solver does not support objective functions")

    def maximize(self, expr):
        """
            Maximize the given objective function

            `maximize()` can be called multiple times, only the last one is stored
        """
        raise NotImplementedError("Solver does not support objective functions")

    def status(self):
        return self.cpm_status

    def solve(self, model, time_limit=None):
        """
            Build the CPMpy model into solver-supported model ready for solving
            and returns the answer (True/False/objective.value())

            Overwrites self.cpm_status

        :param model: CPMpy model to be parsed.
        :type model: Model

        :param time_limit: optional, time limit in seconds
        :type time_limit: int or float

        :return: Bool or Int
        """
        return False

    # OPTIONAL functions

    def solution_hint(self, cpm_vars, vals):
        """
        For warmstarting the solver with a variable assignment

        Typically implemented in SAT-based solvers

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """
        raise NotImplementedError("Solver does not support solution hinting")

    def get_core(self):
        """
        For use with s.solve(assumptions=[...]). Only meaningful if the solver returned UNSAT.

        Typically implemented in SAT-based solvers
        
        Returns a small subset of assumption literals that are unsat together.
        (a literal is either a `_BoolVarImpl` or a `NegBoolView` in case of its negation, e.g. x or ~x)
        Setting these literals to True makes the model UNSAT, setting any to False makes it SAT
        """
        raise NotImplementedError("Solver does not support unsat core extraction")


    # shared helper functions

    def _solve_return(self, cpm_status, objective_value):
        """
            Take a CPMpy Model and SolverStatus object and return
            the proper answer (True/False/objective_value)

        :param cpm_status: status extracted from the solver
        :type cpm_status: SolverStatus

        :param objective_value: None or Int, as computed by solver

        :return: Bool or Int
        """
        # return computed value
        if objective_value is not None and \
            (cpm_status.exitstatus == ExitStatus.OPTIMAL or \
             cpm_status.exitstatus == ExitStatus.FEASIBLE):
            # optimisation problem
            return objective_value
        else:
            # satisfaction problem
            if cpm_status.exitstatus == ExitStatus.FEASIBLE or \
               cpm_status.exitstatus == ExitStatus.OPTIMAL:
                return True
        return False


#==============================================================================
class ExitStatus(Enum):
    """
    Exit status of the solver

    Attributes:

        `NOT_RUN`: Has not been run

        `OPTIMAL`: Optimal solution to an optimisation problem found

        `FEASIBLE`: Feasible solution to a satisfaction problem found,
                    or feasible (but not proven optimal) solution to an
                    optimisation problem found

        `UNSATISFIABLE`: No satisfying solution exists

        `ERROR`: Some error occured (solver should have thrown Exception)

        `UNKNOWN`: Outcome unknown, for example when timeout is reached
    """
    NOT_RUN = 1
    OPTIMAL = 2
    FEASIBLE = 3
    UNSATISFIABLE = 4
    ERROR = 5
    UNKNOWN = 6

#==============================================================================
class SolverStatus(object):
    """
        Status and statistics of a solver run
    """
    exitstatus: ExitStatus
    runtime: time

    def __init__(self, name):
        self.solver_name = name
        self.exitstatus = ExitStatus.NOT_RUN
        self.runtime = None

    def __repr__(self):
        return "{} ({} seconds)".format(self.exitstatus, self.runtime)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## utils.py
##
"""
    Utilities for handling solvers

    Contains a static variable `builtin_solvers` that lists
    CPMpy solvers (first one is the default solver by default)

    =================
    List of functions
    =================

    .. autosummary::
        :nosignatures:

        get_supported_solvers
        param_combinations
"""

#from .minizinc import CPMpyMiniZinc # closed for maintenance
from .ortools import CPM_ortools
from .pysat import CPM_pysat

def get_supported_solvers():
    """
        Returns a list of solvers supported on this machine.

    :return: a list of SolverInterface sub-classes :list[SolverInterface]:
    """
    return [sv for sv in builtin_solvers if sv.supported()]

def param_combinations(all_params, remaining_keys=None, cur_params=None):
    """
        Recursively yield all combinations of param values

        For example usage, see `examples/advanced/hyperparameter_search.py`

        - all_params is a dict of {key: list} items, e.g.:
            {'val': [1,2], 'opt': [True,False]}

        - output is an generator over all {key:value} combinations
          of the keys and values. For the example above:
          generator([{'val':1,'opt':True},{'val':1,'opt':False},{'val':2,'opt':True},{'val':2,'opt':False}])
    """
    if remaining_keys is None or cur_params is None:
        # init
        remaining_keys = list(all_params.keys())
        cur_params = dict()

    cur_key = remaining_keys[0]
    myresults = [] # (runtime, cur_params)
    for cur_value in all_params[cur_key]:
        cur_params[cur_key] = cur_value
        if len(remaining_keys) == 1:
            # terminal, return copy
            yield dict(cur_params)
        else:
            # recursive call
            yield from param_combinations(all_params, 
                            remaining_keys=remaining_keys[1:],
                            cur_params=cur_params)


# Order matters! first is default, then tries second, etc...
builtin_solvers=[CPM_ortools,CPM_pysat]



"""
    All forms of expression objects that allow you to specify constraints and objectives over variables

    ==================
    List of submodules
    ==================
    .. autosummary::
        :nosignatures:

        variables
        core
        globalconstraints
        python_builtins
        utils


"""

# we only import methods/classes that are used for modelling
# others need to be imported by the developer explicitely
from .variables import boolvar, intvar, cpm_array
from .variables import BoolVar, IntVar, cparray # Old, to be deprecated
from .globalconstraints import AllDifferent, AllEqual, Circuit, Table, Minimum, Maximum, Element
from .globalconstraints import alldifferent, allequal, circuit # Old, to be deprecated
from .python_builtins import all, any, max, min



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## expressions.py
##
"""
    The `Expression` superclass and common subclasses `Expression` and `Operator`.
    
    None of these objects should be directly created, they are automatically created through operator
    overloading on variables and expressions.

    Here is a list of standard python operators and what object (with what expr.name) it creates:

    Comparisons:

    - x == y        Comparison("==", x, y)
    - x != y        Comparison("!=", x, y)
    - x < y         Comparison("<", x, y)
    - x <= y        Comparison("<=", x, y)
    - x > y         Comparison(">", x, y)
    - x >= y        Comparison(">=", x, y)

    Mathematical operators:

    - -x            Operator("-", [x])
    - abs(x)        Operator("abs", [x])
    - x + y         Operator("sum", [x,y])
    - sum([x,y,z])  Operator("sum", [x,y,z])
    - x - y         Operator("sum", [x,-y])
    - x * y         Operator("mul", [x,y])
    - x / y         Operator("div", [x,y])
    - x % y         Operator("mod", [x,y])
    - x ** y        Operator("pow", [x,y])

    Logical operators:

    - x & y         Operator("and", [x,y])
    - x | y         Operator("or", [x,y])
    - x ^ y         Operator("xor", [x,y])

    Finally there are two special cases for logical operators 'implies' and '~/not'.
    
    Python has no built-in operator for __implication__ that can be overloaded.
    CPMpy hence has a function 'implies()' that can be called:

    - x.implies(y)  Operator("->", [x,y])

    For negation, we rewrite this to the more generic expression `x == 0`.
    (which in turn creates a `NegBoolView()` in case x is a Boolean variable)

    - ~x            x == 0


    Apart from operator overleading, expressions implement two important functions:

    - `is_bool()`   which returns whether the __return type__ of the expression is Boolean.
                    If it does, the expression can be used as top-level constraint
                    or in logical operators.

    - `value()`     computes the value of this expression, by calling .value() on its
                    subexpressions and doing the appropriate computation
                    this is used to conveniently print variable values, objective values
                    and any other expression value (e.g. during debugging).
    
    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        Expression
        Comparison
        Operator
"""
from types import GeneratorType
from collections.abc import Iterable
import numpy as np
from .utils import is_num, is_any_list, flatlist

class Expression(object):
    """
    An Expression represents a symbolic function with a self.name and self.args (arguments)

    Each Expression is considered to be a function whose value can be used
      in other expressions

    Expressions may implement:
    - is_bool():    whether its return type is Boolean
    - value():      the value of the expression, default None
    - implies(x):   logical implication of this expression towards x
    - __repr__():   for pretty printing the expression
    - any __op__ python operator overloading
    """

    def __init__(self, name, arg_list):
        self.name = name

        if isinstance(arg_list, (tuple,GeneratorType)):
            arg_list = list(arg_list)
        elif isinstance(arg_list, np.ndarray):
            # must flatten
            arg_list = arg_list.reshape(-1)
        for i in range(len(arg_list)):
            if isinstance(arg_list[i], np.ndarray):
                # must flatten
                arg_list[i] = arg_list[i].reshape(-1)

        assert (is_any_list(arg_list)), "_list_ of arguments required, even if of length one e.g. [arg]"
        self.args = arg_list


    def __repr__(self):
        strargs = []
        for arg in self.args:
            if isinstance(arg, np.ndarray):
                # flatten
                strarg = ",".join(map(str,arg.flat))
                strargs.append( f"[{strarg}]" )
            else:
                strargs.append( f"{arg}" )
        return "{}({})".format(self.name, ",".join(strargs))

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
            Default: yes
        """
        return True

    def value(self):
        return None # default

    # implication constraint: self -> other
    # Python does not offer relevant syntax...
    # for double implication, use equivalence self == other
    def implies(self, other):
        # other constant
        if other is True:
            return True
        if other is False:
            return ~self
        return Operator('->', [self, other])

    # Comparisons
    def __eq__(self, other):
        return Comparison("==", self, other)
    def __ne__(self, other):
        return Comparison("!=", self, other)
    def __lt__(self, other):
        return Comparison("<", self, other)
    def __le__(self, other):
        return Comparison("<=", self, other)
    def __gt__(self, other):
        return Comparison(">", self, other)
    def __ge__(self, other):
        return Comparison(">=", self, other)

    # Boolean Operators
    # Implements bitwise operations & | ^ and ~ (and, or, xor, not)
    def __and__(self, other):
        # some simple constant removal
        if other is True:
            return self
        if other is False:
            return False

        return Operator("and", [self, other])
    def __rand__(self, other):
        # some simple constant removal
        if other is True:
            return self
        if other is False:
            return False

        return Operator("and", [other, self])

    def __or__(self, other):
        # some simple constant removal
        if other is True:
            return True
        if other is False:
            return self

        return Operator("or", [self, other])
    def __ror__(self, other):
        # some simple constant removal
        if other is True:
            return True
        if other is False:
            return self

        return Operator("or", [other, self])

    def __xor__(self, other):
        # some simple constant removal
        if other is True:
            return ~self
        if other is False:
            return self

        return Operator("xor", [self, other])
    def __rxor__(self, other):
        # some simple constant removal
        if other is True:
            return ~self
        if other is False:
            return self

        return Operator("xor", [other, self])

    # Mathematical Operators, including 'r'everse if it exists
    # Addition
    def __add__(self, other):
        if is_num(other) and other == 0:
            return self
        return Operator("sum", [self, other])
    def __radd__(self, other):
        if is_num(other) and other == 0:
            return self
        return Operator("sum", [other, self])

    # substraction
    def __sub__(self, other):
        # if is_num(other) and other == 0:
        #     return self
        # return Operator("sub", [self, other])
        return self.__add__(-other)
    def __rsub__(self, other):
        # if is_num(other) and other == 0:
        #     return -self
        # return Operator("sub", [other, self])
        return (-self).__radd__(other)
    
    # multiplication, puts the 'constant' (other) first
    def __mul__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("mul", [self, other])
    def __rmul__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("mul", [other, self])

    # matrix multipliciation TODO?
    #object.__matmul__(self, other)

    # other mathematical ones
    def __truediv__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("div", [self, other])
    def __rtruediv__(self, other):
        return Operator("div", [other, self])

    def __mod__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("mod", [self, other])
    def __rmod__(self, other):
        return Operator("mod", [other, self])

    def __pow__(self, other, modulo=None):
        assert (modulo is None), "Power operator: modulo not supported"
        return Operator("pow", [self, other])
    def __rpow__(self, other, modulo=None):
        assert (modulo is None), "Power operator: modulo not supported"
        return Operator("pow", [other, self])

    # Not implemented: (yet?)
    #object.__floordiv__(self, other)
    #object.__divmod__(self, other)

    # unary mathematical operators
    def __neg__(self):
        return Operator("-", [self])
    def __pos__(self):
        return self
    def __abs__(self):
        return Operator("abs", [self])
    # 'not' for now, no unary constraint for it
    def __invert__(self):
        return (self == 0)


class Comparison(Expression):
    """Represents a comparison between two sub-expressions
    """
    allowed = {'==', '!=', '<=', '<', '>=', '>'}

    def __init__(self, name, left, right):
        assert (name in Comparison.allowed), "Symbol not allowed"
        super().__init__(name, [left, right])

    def __repr__(self):
        if all(isinstance(x, Expression) for x in self.args):
            return "({}) {} ({})".format(self.args[0], self.name, self.args[1]) 
        # if not: prettier printing without braces
        return "{} {} {}".format(self.args[0], self.name, self.args[1]) 

    # is bool, check special case
    def __eq__(self, other):
        if is_num(other) and other == 1:
            return self
        return super().__eq__(other)
        
    # return the value of the expression
    # optional, default: None
    def value(self):
        arg_vals = [arg.value() if isinstance(arg, Expression) else arg for arg in self.args]
        if any(a is None for a in arg_vals): return None
        if   self.name == "==": return (arg_vals[0] == arg_vals[1])
        elif self.name == "!=": return (arg_vals[0] != arg_vals[1])
        elif self.name == "<":  return (arg_vals[0] < arg_vals[1])
        elif self.name == "<=": return (arg_vals[0] <= arg_vals[1])
        elif self.name == ">":  return (arg_vals[0] > arg_vals[1])
        elif self.name == ">=": return (arg_vals[0] >= arg_vals[1])
        return None # default


class Operator(Expression):
    """
    All kinds of mathematical and logical operators on expressions

    Convention for 2-ary operators: if one of the two is a constant,
    it is stored first (as expr[0]), this eases weighted sum detection
    """
    allowed = {
        #name: (arity, is_bool)       arity 0 = n-ary, min 2
        'and': (0, True),
        'or':  (0, True),
        'xor': (0, True),
        '->':  (2, True),
        'sum': (0, False),
        'sub': (2, False), # x - y
        'mul': (2, False),
        'div': (2, False),
        'mod': (2, False),
        'pow': (2, False),
        '-':   (1, False), # -x
        'abs': (1, False),
    }
    printmap = {'sum': '+', 'sub': '-', 'mul': '*', 'div': '/'}

    def __init__(self, name, arg_list):
        # sanity checks
        assert (name in Operator.allowed), "Operator {} not allowed".format(name)
        arity, is_bool = Operator.allowed[name]
        if arity == 0:
            assert (len(arg_list) >= 2), "Operator: n-ary operators require at least two arguments"
            arg_list = flatlist(arg_list)
        else:
            assert (len(arg_list) == arity), "Operator: {}, number of arguments must be {}".format(name, arity)

        # convention for commutative binary operators:
        # swap if right is constant and left is not
        if len(arg_list) == 2 and is_num(arg_list[1]) and \
           name in {'sum', 'mul', 'and', 'or', 'xor'}:
            arg_list[0], arg_list[1] = arg_list[1], arg_list[0]

        super().__init__(name, arg_list)

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return Operator.allowed[self.name][1]
    
    def __repr__(self):
        printname = self.name
        if printname in Operator.printmap:
            printname = Operator.printmap[printname]

        # special cases
        if self.name == '-': # unary -
            return "-({})".format(self.args[0])

        # infix printing of two arguments
        if len(self.args) == 2:
            # bracketed printing of non-constants
            def wrap_bracket(arg):
                if isinstance(arg, Expression):
                    return f"({arg})"
                return arg
            return "{} {} {}".format(wrap_bracket(self.args[0]),
                                     printname,
                                     wrap_bracket(self.args[1]))

        return "{}({})".format(self.name, self.args)

    # associative operations {'and', 'or', 'xor', 'sum', 'mul'} are chained
    # + special case for weighted sum (sum of mul)

    def __and__(self, other):
        if self.name == 'and':
            self.args.append(other)
            return self
        return super().__and__(other)
    def __rand__(self, other):
        if self.name == 'and':
            self.args.insert(0,other)
            return self
        return super().__rand__(other)

    def __or__(self, other):
        if self.name == 'or':
            self.args.append(other)
            return self
        return super().__or__(other)
    def __ror__(self, other):
        if self.name == 'or':
            self.args.insert(0,other)
            return self
        return super().__ror__(other)

    def __xor__(self, other):
        if self.name == 'xor':
            self.args.append(other)
            return self
        return super().__xor__(other)
    def __rxor__(self, other):
        if self.name == 'xor':
            self.args.insert(0,other)
            return self
        return super().__rxor__(other)

    def __add__(self, other):
        if is_num(other) and other == 0:
            return self

        if self.name == 'sum':
            if not isinstance(other, Iterable):
                self.args.append(other)
            else: # vector
                self.args.extend(other)
            return self
        return super().__add__(other)

    def __radd__(self, other):
        # only for constants
        if is_num(other) and other == 0:
            return self

        if self.name == 'sum':
            if not isinstance(other, Iterable):
                self.args.insert(0, other)
            else: # vector
                self.args[:0] = other # prepend an array
            return self
        return super().__radd__(other)

    # substraction; in case of sum or wsum, add
    def __sub__(self, other):
        if is_num(other) and other == 0:
            return self

        if self.name == 'sum':
            self.args.append(-other)
            return self
        return super().__sub__(other)
    def __rsub__(self, other):
        if is_num(other) and other == 0:
            return -self

        return (-self).__radd__(other)

    # is bool, check special case
    def __eq__(self, other):
        if is_num(other) and other == 1:
            # check if bool operator, do not add == 1
            arity, is_bool = Operator.allowed[self.name]
            if is_bool:
                return self
        return super().__eq__(other)

    def value(self):
        arg_vals = [arg.value() if isinstance(arg, Expression) else arg for arg in self.args]
        if any(a is None for a in arg_vals): return None
        if   self.name == "sum": return sum(arg_vals)
        elif self.name == "mul": return arg_vals[0] * arg_vals[1]
        elif self.name == "sub": return arg_vals[0] - arg_vals[1]
        elif self.name == "div": return arg_vals[0] / arg_vals[1]
        elif self.name == "mod": return arg_vals[0] % arg_vals[1]
        elif self.name == "pow": return arg_vals[0] ** arg_vals[1]
        elif self.name == "-":   return -arg_vals[0]
        elif self.name == "abs": return -arg_vals[0] if arg_vals[0] < 0 else arg_vals[0]
        return None # default




#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## pysat.py
##
"""
    Interface to PySAT's API

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_pysat
"""
from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions.core import *
from ..expressions.variables import _BoolVarImpl, NegBoolView
from ..expressions.utils import is_any_list
from ..transformations.get_variables import get_variables_model
from ..transformations.flatten_model import flatten_constraint

class CPM_pysat(SolverInterface):
    """
    Interface to PySAT's API

    Requires that the 'python-sat' python package is installed:
    $ pip install python-sat

    See detailed installation instructions at:
    https://pysathq.github.io/installation.html

    Creates the following attributes:
    user_vars: variables in the original (unflattened) model (for reverse mapping the values after solve)
    pysat_vpool: a pysat.formula.IDPool for the variable mapping
    pysat_solver: a pysat.solver.Solver() (default: glucose4)
    cpm_status: the corresponding CPMpy status
    """

    @staticmethod
    def supported():
        try:
            import pysat
            return True
        except ImportError as e:
            return False

    def __init__(self, cpm_model):
        """
        Constructor of the solver object

        Requires a CPMpy model as input, and will create the corresponding
        PySAT clauses and solver object

        WARNING: CPMpy uses 'model' to refer to a constraint specification,
        the PySAT docs use 'model' to refer to a solution.

        Only supports satisfaction problems (no objective)
        Supports only clauses at this moment!!
        Support for conjunctions and implications can be added through
        translation to 'aiger', see PySAT's documentation
        """
        if not self.supported():
            raise Exception("CPM_pysat: Install the python 'python-sat' package to use this solver interface")
        if cpm_model.objective is not None:
            raise Exception("CPM_pysat: only satisfaction, does not support an objective function")
        from pysat.formula import IDPool
        from pysat.solvers import Solver

        super().__init__()
        self.name = "pysat"

        # store original vars
        self.user_vars = get_variables_model(cpm_model)

        # ID pool of variables
        self.pysat_vpool = IDPool()

        # create constraint model (list of clauses)
        cnf = self.make_cnf(cpm_model)
        # create the solver instance
        self.pysat_solver = Solver(bootstrap_with=cnf.clauses, use_timer=True)


    def pysat_var(self, cpm_var):
        """
            Transforms cpm_var into CNF literal using self.pysat_vpool
            (positive or negative integer)
        """
        if isinstance(cpm_var, NegBoolView):
            # just a view, get actual var identifier, return -id
            return -self.pysat_vpool.id(cpm_var._bv.name)
        elif isinstance(cpm_var, _BoolVarImpl):
            return self.pysat_vpool.id(cpm_var.name)
        else:
            raise NotImplementedError(f"CPM_pysat: variable {cpm_var} not supported")

    def __add__(self, cpm_con):
        """
        Direct solver access constraint addition,
        immediately adds the constraint to PySAT

        Note that we don't store the resulting cpm_model, we translate
        directly to the ort_model

        :param cpm_con CPMpy constraint, or list thereof
        :type cpm_con (list of) Expression(s)
        """
        # base case, just var or ~var
        if isinstance(cpm_con, _BoolVarImpl):
            self.pysat_solver.add_clause([ self.pysat_var(cpm_con) ])
        elif isinstance(cpm_con, Operator):
            if cpm_con.name == 'or':
                self.pysat_solver.add_clause([ self.pysat_var(var) for var in cpm_con.args ])
            else:
                raise NotImplementedError("Only 'or' operator supported by CPM_pysat for now (more possible with aiger, contact us on github")
                    
        # generic case, list
        elif is_any_list(cpm_con):
            # call recursively
            for con in cpm_con:
                self += con
        else:
            raise NotImplementedError(f"Non-operator constraint {cpm_con} not supported by CPM_pysat")

        return self


    def solution_hint(self, cpm_vars, vals):
        """
        PySAT supports warmstarting the solver with a feasible solution

        In PySAT, this is called setting the 'phases' or the 'polarities' of literals

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """
        literals = []
        for (cpm_var, val) in zip(cpm_vars, vals):
            lit = self.pysat_var(cpm_var)
            if val:
                # true, so positive literal
                literals.append(lit)
            else:
                # false, so negative literal
                literals.append(-lit)
        self.pysat_solver.set_phases(literals)


    def solve(self, time_limit=None, assumptions=None):
        """
            - assumptions: list of CPMpy Boolean variables that are assumed to be true.
                           For use with s.get_core(): if the model is UNSAT, get_core() returns a small subset of assumption variables that are unsat together.
                           Note: the PySAT interface is statefull, so you can incrementally call solve() with assumptions and it will reuse learned clauses
        """

        # set time limit?
        if time_limit is not None:
            raise NotImplementedError("Didn't get to it yet, see pysat.solver.interrupt() for an example of what to implement")

        if assumptions is None:
            pysat_assum_vars = [] # default if no assumptions
        else:
            pysat_assum_vars = [self.pysat_var(v) for v in assumptions]
            self.assumption_vars = assumptions

        pysat_status = self.pysat_solver.solve(assumptions=pysat_assum_vars)

        return self._after_solve(pysat_status)

    def _after_solve(self, pysat_status):
        """
            To be called immediately after calling pysat solve() or solve_limited()
            Translate pysat status and variable values to CPMpy corresponding things

            - pysat_status: Boolean or None
        """

        # translate exit status
        self.cpm_status = SolverStatus(self.name)
        if pysat_status is True:
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif pysat_status is False:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif pysat_status is None:
            # can happen when timeout is reached...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else: # another?
            raise NotImplementedError(pysat_status) # a new status type was introduced, please report on github

        # translate runtime
        self.cpm_status.runtime = self.pysat_solver.time()

        # translate solution values (of original vars only)
        if self.cpm_status.exitstatus == ExitStatus.FEASIBLE:
            sol = frozenset(self.pysat_solver.get_model()) # to speed up lookup
            # fill in variables
            for cpm_var in self.user_vars:
                lit = self.pysat_var(cpm_var)
                if lit in sol:
                    cpm_var._value = True
                elif -lit in sol:
                    cpm_var._value = False
                else:
                    # not specified...
                    cpm_var._value = None
                    pass

        return self._solve_return(self.cpm_status, pysat_status)


    def get_core(self):
        """
            For use with s.solve(assumptions=[...]). Only meaningful if the solver returned UNSAT. In that case, get_core() returns a small subset of assumption variables that are unsat together.

            CPMpy will return only those variables that are False (in the UNSAT core)

            Note that there is no guarantee that the core is minimal, though this interface does open up the possibility to add more advanced Minimal Unsatisfiabile Subset algorithms on top. All contributions welcome!
        """
        assert hasattr(self, 'assumption_vars'), "get_core(): requires a list of assumption variables, e.g. s.solve(assumptions=[...])"
        assert (self.cpm_status.exitstatus == ExitStatus.UNSATISFIABLE), "get_core(): solver must return UNSAT"

        assum_idx = frozenset(self.pysat_solver.get_core()) # to speed up lookup

        return [v for v in self.assumption_vars if self.pysat_var(v) in assum_idx]

    def make_cnf(self, cpm_model):
        """
            Makes a pysat.formulae CNF out of 
            a CPMpy model (only supports clauses for now)

            Typically only needed for internal use
        """
        from pysat.formula import CNF

        # check only BoolVarImpl (incl. NegBoolView)
        for var in get_variables_model(cpm_model):
            if not isinstance(var, _BoolVarImpl):
                raise NotImplementedError("Non-Boolean variables not (yet) supported. Reach out on github if you want to help implement a translation")

        # CNF object
        cnf = CNF()

        # Post the constraint expressions to the solver
        # only CNF (list of disjunctions) supported for now
        for con in flatten_constraint(cpm_model.constraints):
            # TODO, perhaps we should check for lists of lists, or top-level ands?
            # base case, just var or ~var
            if isinstance(con, _BoolVarImpl):
                cnf.append([ self.pysat_var(con) ])
            elif isinstance(con, Operator):
                if con.name == 'or':
                    cnf.append([ self.pysat_var(var) for var in con.args ])
                else:
                    raise NotImplementedError("Only 'or' operator supported by CPM_pysat for now (more possible with aiger, contact us on github")
                    
            else:
                raise NotImplementedError(f"Non-operator constraint {con} not supported by CPM_pysat")

        return cnf




#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## ortools.py
##
"""
    Interface to ortools' Python API

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_ortools
"""
from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions.core import Expression, Comparison, Operator
from ..expressions.variables import _NumVarImpl, _IntVarImpl, _BoolVarImpl, NegBoolView
from ..expressions.utils import is_num, is_any_list
from ..transformations.get_variables import get_variables_model, get_variables
from ..transformations.flatten_model import flatten_model, flatten_constraint, flatten_objective, get_or_make_var, negated_normal

class CPM_ortools(SolverInterface):
    """
    Interface to the python 'ortools' API

    Requires that the 'ortools' python package is installed:
    $ pip install ortools

    Creates the following attributes:
    user_vars: variables in the original (unflattened) model, incl objective
    ort_model: the ortools.sat.python.cp_model.CpModel() created by _model()
    ort_solver: the ortools cp_model.CpSolver() instance used in solve()
    ort_status: the ortools 'status' instance returned by ort_solver.Solve()
    cpm_status: the corresponding CPMpy status
    """

    @staticmethod
    def supported():
        try:
            import ortools
            return True
        except ImportError as e:
            return False

    def __init__(self, cpm_model=None):
        """
        Constructor of the solver object

        Requires a CPMpy model as input, and will create the corresponding
        or-tools model and solver object (ort_model and ort_solver)

        ort_model and ort_solver can both be modified externally before
        calling solve(), a prime way to use more advanced solver features
        """
        if not self.supported():
            raise Exception("Install the python 'ortools' package to use this '{}' solver interface".format(self.name))
        from ortools.sat.python import cp_model as ort

        super().__init__()
        self.name = "ortools"

        if cpm_model is None:
            self.user_vars = []
            self.ort_model = ort.CpModel()
            self.varmap = dict() # cppy var -> solver var
        else:
            # store original vars and objective (before flattening)
            self.user_vars = get_variables_model(cpm_model)

            # create model (includes conversion to flat normal form)
            self.ort_model = self.make_model(cpm_model)

        # create the solver instance
        # (so its params can still be changed before calling solve)
        self.ort_solver = ort.CpSolver()


    def __add__(self, cons):
        """
        Direct solver access constraint addition,
        avoids having to flatten the model of the constructor again
        when calling s.solve() repeatedly.

        Note that we don't store the resulting cpm_model, we translate
        directly to the ort_model

        :param cpm_cons list of CPMpy constraints
        :type cpm_cons list of Expressions
        """
        # store new user vars
        new_user_vars = get_variables(cons)
        for v in frozenset(new_user_vars)-frozenset(self.user_vars):
            self.user_vars.append(v)

        flat_cons = flatten_constraint(cons)
        # add new (auxiliary) variables
        for var in get_variables(flat_cons):
            if not var in self.varmap:
                self.add_to_varmap(var)
        # add constraints
        for cpm_con in flat_cons:
            self.post_constraint(cpm_con)

        return self

    def minimize(self, expr):
        """
            Minimize the given objective function

            `minimize()` can be called multiple times, only the last one is stored
        """
        # flatten
        (flat_obj, flat_cons) = flatten_objective(expr)

        # add constraints if needed
        if len(flat_cons) != 0:
            self += flat_cons

        obj = self.ort_numexpr(flat_obj)
        self.ort_model.Minimize(obj)

    def maximize(self, expr):
        """
            Maximize the given objective function

            `maximize()` can be called multiple times, only the last one is stored
        """
        # flatten
        (flat_obj, flat_cons) = flatten_objective(expr)

        # add constraints if needed
        if len(flat_cons) != 0:
            self += flat_cons

        obj = self.ort_numexpr(flat_obj)
        self.ort_model.Maximize(obj)


    def solution_hint(self, cpm_vars, vals):
        """
        or-tools supports warmstarting the solver with a feasible solution

        More specifically, it will branch that variable on that value first if possible. This is known as 'phase saving' in the SAT literature, but then extended to integer variables.

        The solution hint does NOT need to satisfy all constraints, it should just provide reasonable default values for the variables. It can decrease solving times substantially, especially when solving a similar model repeatedly

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """
        for (cpm_var, val) in zip(cpm_vars, vals):
            self.ort_model.ClearHints() # because add just appends
            self.ort_model.AddHint(self.ort_var(cpm_var), val)


    def solve(self, time_limit=None, assumptions=None, **kwargs):
        """
            Arguments:
            - time_limit:  maximum solve time in seconds (float, optional)
            - assumptions: list of CPMpy Boolean variables (or their negation) that are assumed to be true.
                           For use with s.get_core(): if the model is UNSAT, get_core() returns a small subset of assumption variables that are unsat together.
                           Note: the or-tools interace is stateless, so you can incrementally call solve() with assumptions, but or-tools will always start from scratch...

            Additional keyword arguments:
            The ortools solver parameters are defined in its 'sat_parameters.proto' description:
            https://github.com/google/or-tools/blob/stable/ortools/sat/sat_parameters.proto

            You can use any of these parameters as keyword argument to `solve()` and they will
            be forwarded to the solver. Examples include:
                - num_search_workers=8          number of parallel workers (default: 1)
                - log_search_progress=True      to log the search process to stdout (default: False)
                - cp_model_presolve=False       to disable presolve (default: True, almost always beneficial)
                - cp_model_probing_level=0      to disable probing (default: 2, also valid: 1, maybe 3, etc...)
                - linearization_level=0         to disable linearisation (default: 1, can also set to 2)
                - optimize_with_core=True       to do max-sat like lowerbound optimisation (default: False)
                - use_branching_in_lp=True      to generate more info in lp propagator (default: False)
                - polish_lp_solution=True       to spend time in lp propagator searching integer values (default: False)
                - symmetry_level=1              only do symmetry breaking in presolve (default: 2, also possible: 0)

            example:
            o.solve(num_search_workers=8, log_search_progress=True)

        """
        from ortools.sat.python import cp_model as ort

        # set time limit?
        if time_limit is not None:
            self.ort_solver.parameters.max_time_in_seconds = float(time_limit)

        if assumptions is not None:
            ort_assum_vars = [self.ort_var(v) for v in assumptions]
            # this is fucked up... the ort_var()'s index does not seem
            # to match ort_model.VarIndexToVarProto(index)...
            # yet, SufficientAssum... will return that index, so keep own map
            #
            # oh, actually... its a bug that I already reported earlier for
            # VarIndexToVarProto(0) and that Laurent then fixed...
            # Until version 8.3 is released, I'm sticking to own dict
            self.assumption_dict = dict( (ort_var.Index(), cpm_var) for (cpm_var, ort_var) in zip(assumptions, ort_assum_vars) )
            self.ort_model.ClearAssumptions() # because add just appends
            self.ort_model.AddAssumptions(ort_assum_vars)
            # workaround for a presolve with assumptions bug in ortools
            # https://github.com/google/or-tools/issues/2649
            self.ort_solver.parameters.keep_all_feasible_solutions_in_presolve = True

        # set additional keyword arguments in sat_parameters.proto
        for (kw, val) in kwargs.items():
            setattr(self.ort_solver.parameters, kw, val)

        ort_status = self.ort_solver.Solve(self.ort_model)

        return self._after_solve(ort_status)

    def _after_solve(self, ort_status):
        """
            To be called immediately after calling or-tools Solve() or SolveWithSolutionCallBack() or SearchForAllSolutions()
            Translate or-tools status, variable values and objective value to CPMpy corresponding things

            - ort_status, an or-tools 'status' value
        """
        from ortools.sat.python import cp_model as ort

        self.ort_status = ort_status

        # translate exit status
        self.cpm_status = SolverStatus(self.name)
        if self.ort_status == ort.FEASIBLE:
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif self.ort_status == ort.OPTIMAL:
            self.cpm_status.exitstatus = ExitStatus.OPTIMAL
        elif self.ort_status == ort.INFEASIBLE:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif self.ort_status == ort.MODEL_INVALID:
            raise Exception("OR-Tools says: model invalid")
        elif self.ort_status == ort.UNKNOWN:
            # can happen when timeout is reached...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else: # another?
            raise NotImplementedError(self.ort_status) # a new status type was introduced, please report on github

        # translate runtime
        self.cpm_status.runtime = self.ort_solver.WallTime()

        # translate solution values (of original vars only)
        if self.ort_status == ort.FEASIBLE or self.ort_status == ort.OPTIMAL:
            # fill in variables
            for var in self.user_vars:
                var._value = self.ort_solver.Value(self.varmap[var])

        # translate objective
        objective_value = None
        if self.ort_model.HasObjective():
            objective_value = self.ort_solver.ObjectiveValue()

        return self._solve_return(self.cpm_status, objective_value)


    def get_core(self):
        from ortools.sat.python import cp_model as ort
        """
            For use with s.solve(assumptions=[...]). Only meaningful if the solver returned UNSAT. In that case, get_core() returns a small subset of assumption variables that are unsat together.

            CPMpy will return only those variables that are False (in the UNSAT core)

            Note that there is no guarantee that the core is minimal, though this interface does upon up the possibility to add more advanced Minimal Unsatisfiabile Subset algorithms on top. All contributions welcome!

            For pure or-tools example, see http://github.com/google/or-tools/blob/master/ortools/sat/samples/assumptions_sample_sat.py

            Requires or-tools >= 8.2!!!
        """
        assert (self.assumption_dict is not None), "get_core(): requires a list of assumption variables, e.g. s.solve(assumptions=[...])"
        assert (self.ort_status == ort.INFEASIBLE), "get_core(): solver must return UNSAT"

        # use our own dict because of VarIndexToVarProto(0) bug in ort 8.2
        assum_idx = self.ort_solver.SufficientAssumptionsForInfeasibility()

        return [self.assumption_dict[i] for i in assum_idx]

    def make_model(self, cpm_model):
        """
            Makes the ortools.sat.python.cp_model formulation out of 
            a CPMpy model (will do flattening and other transformations)

            Typically only needed for internal use
        """
        from ortools.sat.python import cp_model as ort

        # Constraint programming engine
        self.ort_model = ort.CpModel()

        # Transform into flattened model
        flat_model = flatten_model(cpm_model)

        # Create corresponding solver variables
        self.varmap = dict() # cppy var -> solver var
        for var in get_variables_model(flat_model):
            self.add_to_varmap(var)

        # Post the (flat) constraint expressions to the solver
        for con in flat_model.constraints:
            self.post_constraint(con)

        # Post the objective
        if flat_model.objective is None:
            pass # no objective, satisfaction problem
        else:
            obj = self.ort_numexpr(flat_model.objective)
            if flat_model.objective_max:
                self.ort_model.Maximize(obj)
            else:
                self.ort_model.Minimize(obj)

        return self.ort_model


    def add_to_varmap(self, cpm_var):
        """
        Add the CPMpy variable to the 'varmap' mapping,
        which maps CPMpy variables to or-tools variables

        Typically only needed for internal use
        """
        if isinstance(cpm_var, _BoolVarImpl):
            revar = self.ort_model.NewBoolVar(str(cpm_var))
        elif isinstance(cpm_var, _IntVarImpl):
            revar = self.ort_model.NewIntVar(cpm_var.lb, cpm_var.ub, str(cpm_var))
        self.varmap[cpm_var] = revar


    def post_constraint(self, cpm_expr, reifiable=False):
        """
            Constraints are expected to be in 'flat normal form' (see flatten_model.py)

            While the normal form is divided in 'base', 'comparison' and 'reified', we
            here regroup it per CPMpy class

            Returns the posted ortools 'Constraint', so that it can be used in reification
            e.g. self.post_constraint(smth, reifiable=True).onlyEnforceIf(self.ort_var(bvar))
            
            - reifiable: ensures only constraints that support reification are returned

            Typically only needed for internal use
        """
        # Base case: Boolean variable
        if isinstance(cpm_expr, _BoolVarImpl):
            return self.ort_model.AddBoolOr( [self.ort_var(cpm_expr)] )
        
        # Comparisons: including base (vars), numeric comparison and reify/imply comparison
        elif isinstance(cpm_expr, Comparison):
            lhs,rhs = cpm_expr.args

            if isinstance(lhs, _BoolVarImpl) and cpm_expr.name == '==':
                # base: bvar == bvar|const
                lvar,rvar = map(self.ort_var, (lhs,rhs))
                return self.ort_model.Add(lvar == rvar)

            elif lhs.is_bool() and cpm_expr.name == '==':
                assert (not reifiable), "can not reify a reification"
                # reified case: boolexpr == var, split into two implications
                lexpr = cpm_expr.args[0]
                rvar = cpm_expr.args[1]
                # split in boolexpr -> var and var -> boolexpr
                self.post_constraint(lexpr.implies(rvar))
                self.post_constraint(rvar.implies(lexpr))

            else:
                # numeric (non-reify) comparison case
                rvar = self.ort_var(rhs)
                # lhs can be numexpr
                if isinstance(lhs, _NumVarImpl):
                    # simplest LHS case, a var
                    newlhs = self.ort_var(lhs)
                else:
                    if isinstance(lhs, Operator) and (lhs.name == 'sum' or lhs.name == 'wsum'):
                        # a BoundedLinearExpression LHS, special case, like in objective
                        newlhs = self.ort_numexpr(lhs) 
                    elif cpm_expr.name == '==' and not reifiable:
                        newlhs = None
                        if lhs.name == 'abs':
                            return self.ort_model.AddAbsEquality(rvar, self.ort_var(lhs.args[0]))
                        elif lhs.name == 'mul':
                            return self.ort_model.AddMultiplicationEquality(rvar, self.ort_var_or_list(lhs.args))
                        elif lhs.name == 'mod':
                            return self.ort_model.AddModuloEquality(rvar, *self.ort_var_or_list(lhs.args))
                        elif lhs.name == 'div':
                            return self.ort_model.AddDivisionEquality(rvar, *self.ort_var_or_list(lhs.args))
                        elif lhs.name == 'min':
                            return self.ort_model.AddMinEquality(rvar, self.ort_var_or_list(lhs.args))
                        elif lhs.name == 'max':
                            return self.ort_model.AddMaxEquality(rvar, self.ort_var_or_list(lhs.args))
                        elif lhs.name == 'element':
                            # arr[idx]==rvar (arr=arg0,idx=arg1), ort: (idx,arr,target)
                            return self.ort_model.AddElement(self.ort_var(lhs.args[1]), self.ort_var_or_list(lhs.args[0]), rvar)
                        else:
                            raise NotImplementedError("Not a know supported ORTools left-hand-side '{}' {}".format(lhs.name, cpm_expr))
                    else:
                        # other equality than == 
                        # example: x*y > 10 :: x*y == aux, aux > 10
                        # creat the equality (will handle appropriate bounds)
                        (newvar, cons) = get_or_make_var(lhs)
                        self.add_to_varmap(newvar)
                        for con in cons:
                            # post the flattened constraints, including the 'lhs == newvar' one
                            # if this contains new auxiliary variables we will crash
                            self.post_constraint(con)
                        newlhs = self.ort_var(newvar)

                if newlhs is None:
                    pass # is already posted directly, eg a '=='
                elif cpm_expr.name == '==':
                    return self.ort_model.Add( newlhs == rvar)
                elif cpm_expr.name == '!=':
                    return self.ort_model.Add( newlhs != rvar )
                elif cpm_expr.name == '<=':
                    return self.ort_model.Add( newlhs <= rvar )
                elif cpm_expr.name == '<':
                    return self.ort_model.Add( newlhs < rvar )
                elif cpm_expr.name == '>=':
                    return self.ort_model.Add( newlhs >= rvar )
                elif cpm_expr.name == '>':
                    return self.ort_model.Add( newlhs > rvar )

        # Operators: base (bool), lhs=numexpr, lhs|rhs=boolexpr (reified ->)
        elif isinstance(cpm_expr, Operator):
            if cpm_expr.name == '->' and \
             (not isinstance(cpm_expr.args[0], _BoolVarImpl) or \
              not isinstance(cpm_expr.args[1], _BoolVarImpl)):
                # reified case: var -> boolexpr, boolexpr -> var
                if isinstance(cpm_expr.args[0], _BoolVarImpl):
                    # var -> boolexpr, natively supported by or-tools
                    bvar = self.ort_var(cpm_expr.args[0])
                    return self.post_constraint(cpm_expr.args[1], reifiable=True).OnlyEnforceIf(bvar)
                else:
                    # boolexpr -> var, have to convert to ~var -> ~boolexpr
                    negbvar = self.ort_var(cpm_expr.args[1]).Not()
                    negleft = negated_normal(cpm_expr.args[0])
                    return self.post_constraint(negleft, reifiable=True).OnlyEnforceIf(negbvar)

            else:
                # base 'and'/n, 'or'/n, 'xor'/n, '->'/2
                args = [self.ort_var(v) for v in cpm_expr.args]

                if cpm_expr.name == 'and':
                    return self.ort_model.AddBoolAnd(args)
                elif cpm_expr.name == 'or':
                    return self.ort_model.AddBoolOr(args)
                elif cpm_expr.name == 'xor':
                    return self.ort_model.AddBoolXor(args)
                elif cpm_expr.name == '->':
                    return self.ort_model.AddImplication(args[0],args[1])
                else:
                    raise NotImplementedError("Not a know supported ORTools Operator '{}' {}".format(cpm_expr.name, cpm_expr))

        # rest: base (Boolean) global constraints
        else:
            args = [self.ort_var_or_list(v) for v in cpm_expr.args]

            if cpm_expr.name == 'alldifferent':
                return self.ort_model.AddAllDifferent(args) 
            elif cpm_expr.name == 'table':
                assert(len(args) == 2) # args = [array, table]
                return self.ort_model.AddAllowedAssignments(args[0], args[1])
            # TODO: NOT YET MAPPED: Automaton, Circuit, Cumulative,
            #    ForbiddenAssignments, Inverse?, NoOverlap, NoOverlap2D,
            #    ReservoirConstraint, ReservoirConstraintWithActive
            else:
                # global constraint not known, try generic decomposition
                dec = cpm_expr.decompose()
                if not dec is None:
                    flatdec = flatten_constraint(dec)

                    # collect and create new variables
                    for var in get_variables(flatdec):
                        if not var in self.varmap:
                            self.add_to_varmap(var)
                    # post decomposition
                    for con in flatdec:
                        self.post_constraint(con)
                    # XXX how to deal with reification of such a global??
                    # TODO: we would have to catch this at the time of the reification... (outer call)
                    return None # will throw error if used in reification...
                else:
                    raise NotImplementedError(cpm_expr) # if you reach this... please report on github

    def ort_var(self, cpm_var):
        """
            Uses 'varmap' to return the corresponding or-tools variable
            (or a constant)

            Typically only needed for internal use
        """
        if is_num(cpm_var):
            return cpm_var

        # decision variables, check in varmap
        if isinstance(cpm_var, NegBoolView):
            return self.varmap[cpm_var._bv].Not()
        elif isinstance(cpm_var, _NumVarImpl): # _BoolVarImpl is subclass of _NumVarImpl
            return self.varmap[cpm_var]

        raise NotImplementedError("Not a know var {}".format(cpm_var))

    def ort_var_or_list(self, cpm_expr):
        """
            like ort_var() but also works on lists of variables

            Typically only needed for internal use
        """
        if is_any_list(cpm_expr):
            return [self.ort_var_or_list(sub) for sub in cpm_expr]
        return self.ort_var(cpm_expr)


    def ort_numexpr(self, cpm_expr):
        """
            converts CPMpy subexpression into corresponding
            ORTools subexpressions (for in objective function and comparisons)
            Accepted by ORTools:
            - Decision variable: Var
            - Linear: sum([Var])                                   (CPMpy class 'Operator', name 'sum')
                      wsum([Const],[Var])                          (CPMpy class 'Operator', name 'wsum')

            Typically only needed for internal use
        """
        if is_num(cpm_expr):
            return cpm_expr

        # decision variables, check in varmap
        if isinstance(cpm_expr, _NumVarImpl): # _BoolVarImpl is subclass of _NumVarImpl
            return self.ort_var(cpm_expr)

        # sum or (to be implemented: wsum)
        if isinstance(cpm_expr, Operator):
            args = [self.ort_var(v) for v in cpm_expr.args]
            if cpm_expr.name == 'sum':
                return sum(args) # OR-Tools supports this

        raise NotImplementedError("Not a know supported ORTools expression {}".format(cpm_expr))



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## minizinc.py
##
"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        MiniZincPython

    ==================
    Module description
    ==================

    ==============
    Module details
    ==============
"""

from .solver_interface import ExitStatus, SolverStatus
from .minizinc_text import MiniZincText
from ..transformations.get_variables import get_variables

class CPMpyMiniZinc(MiniZincText):
    """
    Interface to the python 'minizinc' package

    Requires that the 'minizinc' python package is installed:

        $ pip install minizinc
    
    as well as the MiniZinc bundled binary packages, downloadable from:
    https://www.minizinc.org/software.html

    Creates the following attributes:
    mzn_inst: the minizinc.Instance created by _model()
    mzn_result: the minizinc.Result used in solve()
    """

    @staticmethod
    def supported():
        try:
            import minizinc
            return True
        except ImportError as e:
            return False

    def __init__(self):
        self.name = "minizinc_python"
        raise NotImplementedError("Closed for maintenance, has _not yet_ caught up with CPMpy API changes; report in github if this disrupts your plans")

    def _model(self, model, solvername=None):
        import minizinc
        if solvername is None:
            # default solver
            solvername = "gecode"

        # from superclass
        mzn_txt = self.convert(model)

        # minizinc-python API
        # Create a MiniZinc model
        mznmodel = minizinc.Model()
        mznmodel.add_string(mzn_txt)

        # Transform Model into a instance
        slv = minizinc.Solver.lookup(solvername)
        return minizinc.Instance(slv, mznmodel)

    def solve(self, model, solvername=None):
        if not self.supported():
            raise Exception("Install the python 'minizinc' package to use this '{}' solver interface".format(self.name))
        import minizinc

        # create self.mzn_inst
        self.mzn_inst = self._model(model, solvername=solvername)

        # Solve the instance
        self.mzn_result = self.mzn_inst.solve(**{'output-time':True})#all_solutions=True)

        # translate status
        my_status = SolverStatus()
        my_status.solver_name = self.name
        if self.mzn_result.status == minizinc.result.Status.SATISFIED:
            my_status.exitstatus = ExitStatus.FEASIBLE
        elif self.mzn_result.status == minizinc.result.Status.ALL_SOLUTIONS:
            my_status.exitstatus = ExitStatus.FEASIBLE
        elif self.mzn_result.status == minizinc.result.Status.OPTIMAL_SOLUTION:
            my_status.exitstatus = ExitStatus.OPTIMAL
        elif self.mzn_result.status == minizinc.result.Status.UNSATISFIABLE:
            my_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif self.mzn_result.status == minizinc.result.Status.ERROR:
            my_status.exitstatus = ExitStatus.ERROR
            raise Exception("MiniZinc solver returned with status 'Error'")
        elif self.mzn_result.status == minizinc.result.Status.UNKNOWN:
            # means, no solution was found (e.g. within timeout?)...
            my_status.exitstatus = ExitStatus.ERROR
        else:
            raise NotImplementedError # a new status type was introduced, please report on github


        # get runtime and solution
        if 'time' in self.mzn_result.statistics:
            my_status.runtime = self.mzn_result.statistics['time'] # --output-time

        if self.mzn_result.status.has_solution():
            # runtime
            mznsol = self.mzn_result.solution
            my_status.runtime = self.mzn_result.statistics['time'].total_seconds()

            # fill in variables
            modelvars = get_variables(model)
            for var in modelvars:
                varname = str(var)
                if hasattr(mznsol, varname):
                    var._value = getattr(mznsol, varname)
                else:
                    print("Warning, no value for ",varname)

        return my_status



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## model.py
##
"""
    The `Model` class is a lazy container for constraints and an objective function.

    It is lazy in that it only stores the constraints and objective that are added
    to it. Processing only starts when solve() is called, and this does not modify
    the constraints or objective stored in the model.

    A model can be solved multiple times, and constraints can be added to it inbetween
    solve calls.

    See the examples for basic usage, which involves:

    - creation, e.g. m = Model(cons, minimize=obj)
    - solving, e.g. m.solve()
    - optionally, checking status/runtime, e.g. m.status()

    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        Model
"""
import numpy as np
from .expressions.core import Operator
from .expressions.utils import is_any_list
from .solvers.utils import get_supported_solvers
from .solvers.solver_interface import SolverInterface, SolverStatus, ExitStatus

class Model(object):
    """
    CPMpy Model object, contains the constraint and objective expressions
    """

    def __init__(self, *args, minimize=None, maximize=None):
        """
            Arguments of constructor:

            - `*args`: Expression object(s) or list(s) of Expression objects
            - `minimize`: Expression object representing the objective to minimize
            - `maximize`: Expression object representing the objective to maximize

            At most one of minimize/maximize can be set, if none are set, it is assumed to be a satisfaction problem
        """
        assert ((minimize is None) or (maximize is None)), "can not set both minimize and maximize"
        self.cpm_status = SolverStatus("Model") # status of solving this model, will be replaced

        # list of constraints
        if len(args) == 0:
            self.constraints = []
        elif len(args) == 1 and is_any_list(args[0]):
            # top level list of constraints
            self.constraints = args[0]
        else:
            self.constraints = list(args) # instead of tuple

        # objective: an expresion or None
        self.objective = None
        self.objective_max = None
        if maximize is not None:
            self.maximize(maximize)
        if minimize is not None:
            self.minimize(minimize)
        
    def __add__(self, con):
        """
            Add one or more constraints to the model

            m = Model()
            m += [x > 0]
        """
        if is_any_list(con) and len(con) == 1 and is_any_list(con[0]):
            # top level list of constraints
            con = con[0]
        self.constraints.append(con)
        return self

    def minimize(self, expr):
        """
            Minimize the given objective function

            `minimize()` can be called multiple times, only the last one is stored
        """
        self.objective = expr
        self.objective_max = False

    def maximize(self, expr):
        """
            Maximize the given objective function

            `maximize()` can be called multiple times, only the last one is stored
        """
        self.objective = expr
        self.objective_max = True
    
    # solver: name of supported solver or any SolverInterface object
    def solve(self, solver=None, time_limit=None):
        """ Send the model to a solver and get the result

        :param solver: solver to use (verifies that the solver is supported on the current system)
        :type solver: None (default) or in [s.name in get_supported_solvers()] or a SolverInterface class (Class, not object! e.g. CPMpyOrTools, not CPMpyOrTools()!)

        :param time_limit: optional, time limit in seconds
        :type time_limit: int or float

        :return: the computed output:
            - True      if it is a satisfaction problem and it is satisfiable
            - False     if it is a satisfaction problem and not satisfiable
            - [int]     if it is an optimisation problem
        """
        solver_class = None
        # get supported solvers
        supsolvers = get_supported_solvers()
        if solver is None: # default is first
            solver_class = supsolvers[0]
        elif not isinstance(solver, SolverInterface):
            solvername = solver
            for s in supsolvers:
                if s.name == solvername:
                    solver_class = s
                    break # break and hence 'solver' is correct object
        elif isinstance(solver, SolverInterface) and solver.supported():
            solver_class = solver
        else:
            raise Exception("'{}' is not in the list of supported solvers and not a SolverInterface class".format(solver))
        assert(solver_class is not None)
                
        # instatiate solver with this model
        s = solver_class(self)
        # call solver
        ret = s.solve(time_limit=time_limit)
        # store CPMpy status (s object has no further use)
        self.cpm_status = s.status()
        return ret

    def status(self):
        """
            Returns the status of the latest solver run on this model

            Status information includes exit status (optimality) and runtime.

        :return: an object of :class:`SolverStatus`
        """
        return self.cpm_status

    def __repr__(self):
        cons_str = ""
        for c in self.constraints:
            cons_str += "    {}\n".format(c)

        obj_str = ""
        if not self.objective is None:
            if self.objective_max:
                obj_str = "maximize "
            else:
                obj_str = "minimize "
        obj_str += str(self.objective)
            
        return "Constraints:\n{}Objective: {}".format(cons_str, obj_str)



import sys
import copy
from ..expressions.core import *
from ..expressions.variables import _NumVarImpl, _IntVarImpl, _BoolVarImpl, NegBoolView
from ..expressions.utils import is_num, is_any_list

"""
Flattening a model (or individual constraints) into 'flat normal form'.

In flat normal form, constraints belong to one of three families with all arguments
either constants, variables, list of constants or list of variables, and
some binary constraints have a canonical order of variables.

Furthermore, it is 'negated normal' meaning that the ~ (negation operator) only appears
before a Boolean variable (in CPMpy, absorbed in a 'NegBoolView'),
and it is 'negation normal' meaning that the - (negative operator) only appears before
a constant, that is a - b :: a + -1*b :: wsum([1,-1],[a,b])

The three families of possible constraints are:

Base constraints: (no nesting)
-----------------
    - Boolean variable
    - Boolean operators: and([Var]), or([Var]), xor([Var]) (CPMpy class 'Operator', is_bool())
    - Boolean impliciation: Var -> Var                     (CPMpy class 'Operator', is_bool())
    - Boolean equality: Var == Var                         (CPMpy class 'Comparison')
                        Var == Constant                    (CPMpy class 'Comparison')
    - Global constraint (Boolean): global([Var]*)          (CPMpy class 'GlobalConstraint', is_bool())

Comparison constraints: (up to one nesting on one side)
-----------------------
    - Numeric equality:  Numexpr == Var                    (CPMpy class 'Comparison')
                         Numexpr == Constant               (CPMpy class 'Comparison')
    - Numeric disequality: Numexpr != Var                  (CPMpy class 'Comparison')
                           Numexpr != Constant             (CPMpy class 'Comparison')
    - Numeric inequality (>=,>,<,<=): Numexpr >=< Var      (CPMpy class 'Comparison')

    Numexpr:
        - Operator (non-Boolean) with all args Var/constant (examples: +,*,/,mod,wsum)
                                                           (CPMpy class 'Operator', not is_bool())
        - Global constraint (non-Boolean) (examples: Max,Min,Element)
                                                           (CPMpy class 'GlobalConstraint', not is_bool()))
    wsum: wsum([Const],[Var]) represents sum([Const]*[Var]) # TODO: not implemented yet

Reify/imply constraint: (up to two nestings on one side)
-----------------------
    - Reification (double implication): Boolexpr == Var    (CPMpy class 'Comparison')
    - Implication: Boolexpr -> Var                         (CPMpy class 'Operator', is_bool())
                   Var -> Boolexpr                         (CPMpy class 'Operator', is_bool())

    Boolexpr:
        - Boolean operators: and([Var]), or([Var]), xor([Var]) (CPMpy class 'Operator', is_bool())
        - Boolean equality: Var == Var                         (CPMpy class 'Comparison')
        - Global constraint (Boolean): global([Var]*)          (CPMpy class 'GlobalConstraint', is_bool())
        - Comparison constraint (see above)                    (CPMpy class 'Comparison')
    
    Reification of a comparison is the most complex case as it can allow up to 3 levels of nesting in total, e.g.:
        - (wsum([1,2,3],[IV1,IV2,IV3]) > 5) == BV
        - (IV1 == IV2) == BV
        - (BV1 == BV2) == BV3

Objective: (up to one nesting)
---------
    - Satisfaction problem: None
    - Decision variable: Var
    - Linear: sum([Var])                                   (CPMpy class 'Operator', name 'sum')
              wsum([Const],[Var])                          (CPMpy class 'Operator', name 'wsum')

The output after calling flatten_model() or flatten_constraint() will ONLY contain expressions
of the form specified above.

The flattening does not promise to do common subexpression elimination or to automatically group
commutative expressions (and, or, sum, wsum, ...) but such optimisations should be added later.

TODO: remove zipcycle (no longer needed)
TODO: use normalized_boolexpr when possible in the flatten_cons operator case.
TODO: update behind_the_scenes.rst doc with the new 'flat normal form'
TODO: small optimisations, e.g. and/or chaining (potentially after negation), see test_flatten
"""

from itertools import cycle
def __zipcycle(vars1, vars2):
    v1 = [vars1] if not is_any_list(vars1) else vars1
    v2 = [vars2] if not is_any_list(vars2) else vars2
    return zip(v1, cycle(v2)) if len(v2) < len(v1) else zip(cycle(v1), v2)

def flatten_model(orig_model):
    """
        Receives model, returns new model where every constraint is in 'flat normal form'
    """
    from ..model import Model # otherwise circular dependency...

    # the top-level constraints
    basecons = []
    for con in orig_model.constraints:
        basecons += flatten_constraint(con)

    # the objective
    if orig_model.objective is None:
        return Model(*basecons) # no objective, satisfaction problem
    else:
        (newobj, newcons) = flatten_objective(orig_model.objective)
        basecons += newcons
        if orig_model.objective_max:
            return Model(*basecons, maximize=newobj)
        else:
            return Model(*basecons, minimize=newobj)


def flatten_constraint(expr):
    """
        input is any expression; except is_num(), pure _NumVarImpl,
        or Operator/GlobalConstraint with not is_bool()
        
        output: see definition of 'flat normal form' above.

        it will return 'Exception' if something is not supported
        TODO, what built-in python error is best?
    """
    # base cases
    if isinstance(expr, _BoolVarImpl) or isinstance(expr, bool):
        return [expr]
    elif is_num(expr) or isinstance(expr, _NumVarImpl):
        raise Exception("Numeric constants or numeric variables not allowed as base constraint")

    # recursively flatten list of constraints
    if is_any_list(expr):
        flatcons = [flatten_constraint(e) for e in expr]
        return [c for con in flatcons for c in con]
    # recursively flatten top-level 'and'
    if isinstance(expr, Operator) and expr.name == 'and':
        flatcons = [flatten_constraint(e) for e in expr.args]
        return [c for con in flatcons for c in con]

    assert expr.is_bool(), f"Boolean expressions only in flatten_constraint, `{expr}` not allowed."

    if isinstance(expr, Operator):
        """
            - Base Boolean operators: and([Var]), or([Var]), xor([Var]) (CPMpy class 'Operator', is_bool())
            - Base Boolean impliciation: Var -> Var                     (CPMpy class 'Operator', is_bool())
            - Implication: Boolexpr -> Var                         (CPMpy class 'Operator', is_bool())
                           Var -> Boolexpr                         (CPMpy class 'Operator', is_bool())
        """
        # does not type-check that arguments are bool... Could do now with expr.is_bool()!
        if all(__is_flat_var(arg) for arg in expr.args):
            return [expr]
        elif not expr.name == '->':
            # and, or, xor
            # recursively flatten all children
            flatvars, flatcons = zip(*[get_or_make_var(arg) for arg in expr.args])

            newexpr = Operator(expr.name, flatvars)
            return [newexpr]+[c for con in flatcons for c in con]
        else:
            # ->, allows a boolexpr on one side
            if isinstance(expr.args[0], _BoolVarImpl):
                # LHS is var, ensure RHS is normalized 'Boolexpr'
                lhs = expr.args[0]
                (rhs,flatcons) = normalized_boolexpr(expr.args[1])
            else:
                # make LHS normalized 'Boolexpr', RHS must be a var
                (lhs,lcons) = normalized_boolexpr(expr.args[0])
                (rhs,rcons) = get_or_make_var(expr.args[1])
                flatcons = lcons+rcons

            newexpr = Operator(expr.name, (lhs,rhs))
            return [newexpr]+[c for c in flatcons]

    elif isinstance(expr, Comparison):
        """
    - Base Boolean equality: Var == Var                         (CPMpy class 'Comparison')
                             Var == Constant                    (CPMpy class 'Comparison')
    - Numeric equality:  Numexpr == Var                    (CPMpy class 'Comparison')
                         Numexpr == Constant               (CPMpy class 'Comparison')
    - Numeric disequality: Numexpr != Var                  (CPMpy class 'Comparison')
                           Numexpr != Constant             (CPMpy class 'Comparison')
    - Numeric inequality (>=,>,<,<=,): Numexpr >=< Var     (CPMpy class 'Comparison')
    - Reification (double implication): Boolexpr == Var    (CPMpy class 'Comparison')
        """

        flatcons = []
        # zipcycle: unfolds 'arr1 == arr2' pairwise
        # XXX: zipcycle no longer needed, vectorized now handled at creation level!
        for lexpr, rexpr in __zipcycle(expr.args[0], expr.args[1]):
            if __is_flat_var(lexpr) and __is_flat_var(rexpr):
                flatcons += [Comparison(expr.name, lexpr, rexpr)]
            else:
                # RHS must be var (or const)
                lexpr,rexpr = expr.args
                exprname = expr.name
                # ==,!=: can swap if lhs is var and rhs is not
                # TODO: this is very similar to (part of) normalize_boolexpr??
                # XXX indeed, every case that is not 'reification' can be
                # delegated to normalize_boolexpr... TODO
                if (exprname == '==' or exprname == '!=') and \
                    not __is_flat_var(rexpr) and __is_flat_var(lexpr):
                    (lexpr,rexpr) = (rexpr,lexpr)
                # ensure rhs is var
                (rvar, rcons) = get_or_make_var(rexpr)

                # LHS: check if Boolexpr == smth:
                if (exprname == '==' or exprname == '!=') and lexpr.is_bool():
                    if is_num(rexpr):
                        # BoolExpr == 0|False
                        # special case, handled in normalized_boolexpr()
                        (con, subs) = normalized_boolexpr(expr)
                        flatcons += [con] + subs
                        continue # ready with this one

                    # Reification (double implication): Boolexpr == Var
                    if __is_flat_var(lexpr):
                        (lhs, lcons) = (lexpr, [])
                    else:
                        (lhs, lcons) = normalized_boolexpr(lexpr)
                    if expr.name == '!=':
                        # != not needed, negate RHS variable
                        rhs = ~rvar
                        exprname = '=='
                else:
                    # other cases: LHS is numexpr
                    (lhs, lcons) = normalized_numexpr(lexpr)

                flatcons += [Comparison(exprname, lhs, rvar)]+lcons+rcons

        return flatcons

    else:
        """
    - Global constraint (Boolean): global([Var]*)          (CPMpy class 'GlobalConstraint', is_bool())
        """
        # just recursively flatten args, which can be lists
        if all(__is_flat_var_or_list(arg) for arg in expr.args):
            return [expr]
        else:
            # recursively flatten all children
            flatvars, flatcons = zip(*[get_or_make_var_or_list(arg) for arg in expr.args])

            # take copy, replace args
            newexpr = copy.copy(expr) # shallow or deep? currently shallow
            newexpr.args = flatvars
            return [newexpr]+[c for con in flatcons for c in con]


def flatten_objective(expr):
    """
    - Decision variable: Var
    - Linear: sum([Var])                                   (CPMpy class 'Operator', name 'sum')
              wsum([Const],[Var])                          (CPMpy class 'Operator', name 'wsum')
    """
    if __is_flat_var(expr):
        return (expr, [])

    if isinstance(expr, Operator) and (expr.name == 'sum' or expr.name == 'wsum'):
        if expr.name == 'sum':
            if all(__is_flat_var(arg) for arg in expr.args):
                return (expr, [])
            else:
                # one of the arguments is not flat, flatten all
                flatvars, flatcons = zip(*[get_or_make_var(arg) for arg in expr.args])
                newexpr = Operator(expr.name, flatvars)
                return (newexpr, [c for con in flatcons for c in con])
        elif expr.name == 'wsum':
            raise NotImplementedError(expr) # TODO, wsum
    
    # any other numeric expression
    return get_or_make_var(expr)


def __is_flat_var(arg):
    """ True if the variable is a numeric constant, or a _NumVarImpl (incl subclasses)
    """
    return is_num(arg) or isinstance(arg, _NumVarImpl)

def __is_flat_var_or_list(arg):
    """ True if the variable is a numeric constant, or a _NumVarImpl (incl subclasses)
        or a list of __is_flat_var_or_list
    """
    return is_num(arg) or isinstance(arg, _NumVarImpl) or \
           is_any_list(arg) and all(__is_flat_var_or_list(el) for el in arg)


def get_or_make_var(expr):
    """
        Must return a variable, and list of flat normal constraints
        Determines whether this is a Boolean or Integer variable and returns
        the equivalent of: (var, normalize(expr) == var)
    """
    if __is_flat_var(expr):
        return (expr, [])

    if expr.is_bool():
        # normalize expr into a boolexpr LHS, reify LHS == bvar
        (newexpr, flatcons) = normalized_boolexpr(expr)

        bvar = _BoolVarImpl()
        return (bvar, [newexpr == bvar]+flatcons)

    #else:
    # normalize expr into a numexpr LHS, return LHS == intvar
    # includes estimating appropriate bounds for intvar...

    # special case, -var... 
    # XXX until we do weighted sum, turn into -1*args[0]
    if isinstance(expr, Operator) and expr.name == '-': # unary
        return get_or_make_var(-1*expr.args[0])

    if isinstance(expr, Operator):
        # TODO: more like above, call normalized_numexpr() on expr, then equate...
        flatvars, flatcons = zip(*[get_or_make_var(arg) for arg in expr.args]) # also bool, reified...
        lbs = [var.lb if isinstance(var, _NumVarImpl) else var for var in flatvars]
        ubs = [var.ub if isinstance(var, _NumVarImpl) else var for var in flatvars]

        # TODO: weighted sum
        if expr.name == 'abs': # unary
            ivar = _IntVarImpl(0, ubs[0])
        elif expr.name == 'mul': # binary
            lb = lbs[0] * lbs[1]
            ub = ubs[0] * ubs[1]
            if lb > ub: # a negative nr
                lb,ub = ub,lb
            ivar = _IntVarImpl(lb, ub) 
        elif expr.name == 'div': # binary
            lb = lbs[0] // lbs[1]
            ub = ubs[0] // ubs[1]
            if lb > ub: # a negative nr
                lb,ub = ub,lb
            ivar = _IntVarImpl(lb, ub) 
        elif expr.name == 'mod': # binary 
            ivar = _IntVarImpl(0, ubs[0])
        elif expr.name == 'pow': # binary
            ivar = _IntVarImpl(lbs[0] ** lbs[1], ubs[0] ** ubs[1])
        elif expr.name == 'sum': # n-ary
            ivar = _IntVarImpl(sum(lbs), sum(ubs)) 
        elif expr.is_bool(): # Boolean
            ivar = _BoolVarImpl() # TODO: we can support Bool? check, update docs
        else:
            raise Exception("Operator '{}' not known in get_or_make_var".format(expr.name)) # or bug

        newexpr = (Operator(expr.name, flatvars) == ivar)
        return (ivar, [newexpr]+[c for con in flatcons for c in con])

    else:
        """
        - Global constraint (non-Boolean) (examples: Max,Min,Element)
        """
        # just recursively flatten args, which can be lists
        if all(__is_flat_var_or_list(arg) for arg in expr.args):
            newexpr = expr
            flatcons = []
        else:
            flatvars, flatcons = zip(*[get_or_make_var_or_list(arg) for arg in expr.args]) # also bool, reified...
            #idx, icons = flatten_numexpr(idx)
            #arr, acons = zip(*[flatten_numexpr(e) for e in arr])
            #basecons = icons+[c for con in acons for c in con]

            # take copy, replace args
            newexpr = copy.copy(expr) # shallow or deep? currently shallow
            newexpr.args = flatvars

        # XXX Also, how to get the bounds on the new variable? have the solver handle it?
        # XXX Add to GlobalCons as function? e.g. (lb,ub) = expr.get_bounds()? would also work for Operator...
        ivar = _IntVarImpl(-2147483648, 2147483647) # TODO, this can breaks solvers

        return (ivar, [newexpr == ivar]+[c for con in flatcons for c in con])
    

def get_or_make_var_or_list(expr):
    """ Like get_or_make_var() but also accepts and recursively transforms lists
        Used to convert arguments of globals
    """
    if __is_flat_var_or_list(expr):
        return (expr,[])
    elif is_any_list(expr):
        flatvars, flatcons = zip(*[get_or_make_var(arg) for arg in expr])
        return (flatvars, [c for con in flatcons for c in con])
    else:
        return get_or_make_var(expr)


def normalized_boolexpr(expr):
    """
        all 'flat normal form' Boolean expressions that can be 'reified', meaning that

            - expr == BoolVar
            - expr != BoolVar
            - expr -> BoolVar

        are valid expressions.

        Currently, this is the case for:
        - Boolean operators: and([Var]), or([Var]), xor([Var]) (CPMpy class 'Operator', is_bool())
        - Boolean equality: Var == Var                         (CPMpy class 'Comparison')
        - Global constraint (Boolean): global([Var]*)          (CPMpy class 'GlobalConstraint', is_bool())
        - Comparison constraint (see elsewhere)                (CPMpy class 'Comparison')

        output: (base_expr, base_cons) with:
            base_expr: same as 'expr', but all arguments are variables
            base_cons: list of flat normal constraints
    """
    assert(not __is_flat_var(expr))
    assert(expr.is_bool()) 

    if isinstance(expr, Operator):
        # and, or, xor, ->

        # apply De Morgan's transform for "implies"
        if expr.name == '->':
            # TODO, optimisation if args0 is an 'and'?
            (lhs,lcons) = get_or_make_var(expr.args[0])
            # TODO, optimisation if args1 is an 'or'?
            (rhs,rcons) = get_or_make_var(expr.args[1])
            return ((~lhs | rhs), lcons+rcons)

        if all(__is_flat_var(arg) for arg in expr.args):
            return (expr, [])
        else:
            # one of the arguments is not flat, flatten all
            flatvars, flatcons = zip(*[get_or_make_var(arg) for arg in expr.args])
            newexpr = Operator(expr.name, flatvars)
            return (newexpr, [c for con in flatcons for c in con])

    elif isinstance(expr, Comparison):
        if all(__is_flat_var(arg) for arg in expr.args):
            return (expr, [])
        else:
            # LHS can be numexpr, RHS has to be variable

            # TODO: optimisations that swap directions instead when it can avoid to create vars
            """
            if expr.name == '==' or expr.name == '!=':
                # RHS has to be variable, LHS can be more
                if __is_flat_var(lexpr) and not __is_flat_var(rexpr):
                    # LHS is var and RHS not, swap for new expression
                    lexpr, rexpr = rexpr, lexpr

                if __is_numexpr(lexpr) and __is_numexpr(rexpr):
                    # numeric case
                    (lrich, lcons) = flatten_objective(lexpr)
                    (rvar, rcons) = flatten_numexpr(rexpr)
                else:
                    # Boolean case
                    # make LHS reify_ready, RHS var
                    (lrich, lcons) = reify_ready_boolexpr(lexpr)
                    (rvar, rcons) = flatten_boolexpr(rexpr)
                flatcons += [Comparison(expr.name, lrich, rvar)]+lcons+rcons

            else: # inequalities '<=', '<', '>=', '>'
                newname = expr.name
                # LHS can be linexpr, RHS a var
                if __is_flat_var(lexpr) and not __is_flat_var(rexpr):
                    # LHS is var and RHS not, swap for new expression (incl. operator name)
                    lexpr, rexpr = rexpr, lexpr
                    if   expr.name == "<=": newname = ">="
                    elif expr.name == "<":  newname = ">"
                    elif expr.name == ">=": newname = "<="
                    elif expr.name == ">":  newname = "<"

                # make LHS like objective, RHS var
                (lrich, lcons) = flatten_objective(lexpr)
                (rvar, rcons) = flatten_numexpr(rexpr)
                flatcons += [Comparison(newname, lrich, rvar)]+lcons+rcons
            """

            # XXX This is duplicate code from flatten_constraint!!! (except return)
            # -> move into shared function??? or call this one there?
            # But should support one 'less' level of nesting?

            # RHS must be var (or const)
            lexpr,rexpr = expr.args
            exprname = expr.name
            # ==,!=: can swap if lhs is var and rhs is not
            if (exprname == '==' or exprname == '!=') and \
                not __is_flat_var(rexpr) and __is_flat_var(lexpr):
                (lexpr,rexpr) = (rexpr,lexpr)
            # ensure rhs is var
            (rvar, rcons) = get_or_make_var(rexpr)

            # LHS: check if Boolexpr == smth:
            if (exprname == '==' or exprname == '!=') and lexpr.is_bool():
                if is_num(rexpr):
                    # BoolExpr == 0|False
                    assert(not rexpr) # 'true' is preprocessed away

                    nnexpr = negated_normal(lexpr)
                    return normalized_boolexpr(nnexpr)

                # Reification (double implication): Boolexpr == Var
                (lhs, lcons) = normalized_boolexpr(lexpr)
                if expr.name == '!=':
                    # != not needed, negate RHS variable
                    rhs = ~rvar
                    exprname = '=='
            else:
                # other cases: LHS is numexpr
                (lhs, lcons) = normalized_numexpr(lexpr)

            return (Comparison(exprname, lhs, rvar), lcons+rcons)

    else:
        """
        - Global constraint (Boolean): global([Var]*)          (CPMpy class 'GlobalConstraint', is_bool())
        """
        # XXX literal copy from flatten_cons... (except return)
        # just recursively flatten args, which can be lists
        if all(__is_flat_var_or_list(arg) for arg in expr.args):
            return (expr, [])
        else:
            # recursively flatten all children
            flatvars, flatcons = zip(*[get_or_make_var_or_list(arg) for arg in expr.args])

            # take copy, replace args
            newexpr = copy.copy(expr) # shallow or deep? currently shallow
            newexpr.args = flatvars
            return (newexpr, [c for con in flatcons for c in con])


def normalized_numexpr(expr):
    """
        all 'flat normal form' numeric expressions...

        Currently, this is the case for:
        - Operator (non-Boolean) with all args Var/constant (examples: +,*,/,mod,wsum)
                                                           (CPMpy class 'Operator', not is_bool())
        - Global constraint (non-Boolean) (examples: Max,Min,Element)
                                                           (CPMpy class 'GlobalConstraint', not is_bool()))

        output: (base_expr, base_cons) with:
            base_expr: same as 'expr', but all arguments are variables
            base_cons: list of flat normal constraints
    """
    # XXX a boolexpr is also a valid numexpr... e.g. 30*(iv > 5) + ... see mario obj.
    if __is_flat_var(expr):
        return (expr, [])

    # special case, -var... 
    # XXX until we do weighted sum, turn into -1*args[0]
    if isinstance(expr, Operator) and expr.name == '-': # unary
        return normalized_numexpr(-1*expr.args[0])

    if isinstance(expr, Operator):
        if all(__is_flat_var(arg) for arg in expr.args):
            return (expr, [])
        
        # recursively flatten all children
        flatvars, flatcons = zip(*[get_or_make_var(arg) for arg in expr.args])

        newexpr = Operator(expr.name, flatvars)
        return (newexpr, [c for con in flatcons for c in con])

    else:
        """
        - Global constraint (non-Boolean) (examples: Max,Min,Element)
        """
        # XXX literal copy from flatten_cons... (except return)
        # just recursively flatten args, which can be lists
        if all(__is_flat_var_or_list(arg) for arg in expr.args):
            return (expr, [])
        else:
            # recursively flatten all children
            flatvars, flatcons = zip(*[get_or_make_var_or_list(arg) for arg in expr.args])

            # take copy, replace args
            newexpr = copy.copy(expr) # shallow or deep? currently shallow
            newexpr.args = flatvars
            return (newexpr, [c for con in flatcons for c in con])

    raise Exception("Operator '{}' not allowed as numexpr".format(expr)) # or bug

def negated_normal(expr):
    """
        WORK IN PROGRESS
        Negate 'expr' by pushing the negation down into it and its args

        Comparison: swap comparison sign
        Operator.is_bool(): apply DeMorgan
        Global: should call decompose and negate that?

        This function only ensures 'negated normal' for the top-level
        constraint (negating arguments recursively as needed),
        it does not ensure flatness (except if the input is flat)
    """

    if __is_flat_var(expr):
        return ~expr

    elif isinstance(expr, Comparison):
        newexpr = copy.copy(expr)
        if   expr.name == '==': newexpr.name = '!='
        elif expr.name == '!=': newexpr.name = '=='
        elif expr.name == '<=': newexpr.name = '>'
        elif expr.name == '<':  newexpr.name = '>='
        elif expr.name == '>=': newexpr.name = '<'
        elif expr.name == '>':  newexpr.name = '<='
        return newexpr

    elif isinstance(expr, Operator):
        assert(expr.is_bool())

        if expr.name == 'and':
            return Operator('or', [negated_normal(arg) for arg in expr.args])
        elif expr.name == 'or':
            return Operator('and', [negated_normal(arg) for arg in expr.args])
        elif expr.name == '->':
            return expr.args[0] & ~expr.args[1]
        else:
            #raise NotImplementedError("negate_normal {}".format(expr))
            return expr == 0 # can't do better than this...

    else:
        # global...
        #raise NotImplementedError("negate_normal {}".format(expr))
        return expr == 0 # can't do better than this...

